@@ -9,21 +9,15 @@
 #include "src/string/strstr.h"
 
 #include "src/__support/common.h"
-#include <stddef.h>
+#include "src/string/memory_utils/inline_strstr.h"
 
-namespace __llvm_libc {
+namespace LIBC_NAMESPACE {
 
 // TODO: This is a simple brute force implementation. This can be
 // improved upon using well known string matching algorithms.
 LLVM_LIBC_FUNCTION(char *, strstr, (const char *haystack, const char *needle)) {
-  for (size_t i = 0; haystack[i]; ++i) {
-    size_t j;
-    for (j = 0; haystack[i + j] && haystack[i + j] == needle[j]; ++j)
-      ;
-    if (!needle[j])
-      return const_cast<char *>(haystack + i);
-  }
-  return nullptr;
+  auto comp = [](char l, char r) -> int { return l - r; };
+  return inline_strstr(haystack, needle, comp);
 }
 
-} // namespace __llvm_libc
+} // namespace LIBC_NAMESPACE
